@@ -1,4 +1,4 @@
-"""Some utilities for Hydrodata."""
+"""Some utilities for manipulating GeoSpatial data."""
 import numbers
 from typing import Any, Dict, List, Optional, Tuple, Union
 from warnings import warn
@@ -14,15 +14,17 @@ import rasterio.mask as rio_mask
 import rasterio.transform as rio_transform
 import xarray as xr
 from shapely import ops
-from shapely.geometry import LineString, Point, Polygon, box
+from shapely.geometry import LineString, MultiPolygon, Point, Polygon, box
 
-from .exceptions import InvalidInputType
+from .exceptions import InvalidInputType, InvalidInputValue, MissingAttribute
 
 DEF_CRS = "epsg:4326"
 
 
 def json2geodf(
-    content: Union[List[Dict[str, Any]], Dict[str, Any]], in_crs: str = DEF_CRS, crs: str = DEF_CRS,
+    content: Union[List[Dict[str, Any]], Dict[str, Any]],
+    in_crs: str = DEF_CRS,
+    crs: str = DEF_CRS,
 ) -> gpd.GeoDataFrame:
     """Create GeoDataFrame from (Geo)JSON.
 
@@ -251,9 +253,10 @@ def arcgis2geojson(arcgis: Dict[str, Any], id_attr: Optional[str] = None) -> Dic
 
 def gtiff2xarray(
     r_dict: Dict[str, bytes],
-    geometry: Union[Polygon, Tuple[float, float, float, float]],
+    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
     geo_crs: str,
     ds_dims: Tuple[str, str] = ("y", "x"),
+    driver: str = "GTiff",
 ) -> Union[xr.DataArray, xr.Dataset]:
     """Convert responses from ``pygeoogc.wms_bybox`` to ``xarray.Dataset``.
 
@@ -261,19 +264,25 @@ def gtiff2xarray(
     ----------
     r_dict : dict
         The output of ``wms_bybox`` function.
-    geometry : Polygon or tuple
+    geometry : Polygon, MultiPolygon, or tuple
         The geometry to mask the data that should be in the same CRS as the r_dict.
     geo_crs : str
         The spatial reference of the input geometry.
     ds_dims : tuple of str, optional
         The names of the vertical and horizontal dimensions (in that order)
         of the target dataset, default to ("y", "x").
+    driver : str, optional
+        A GDAL driver for reading the content, defaults to GTiff. A list of the drivers
+        can be found here: https://gdal.org/drivers/raster/index.html
 
     Returns
     -------
     xarray.Dataset or xarray.DataAraay
         The dataset or data array based on the number of variables.
     """
+    if not isinstance(r_dict, dict):
+        raise InvalidInputType("r_dict", "dict", '{"name": Response.content}')
+
     key1 = next(iter(r_dict.keys()))
     if len(r_dict) == 1 and "dd" not in key1:
         r_dict = {f"{key1}_dd_0_0": r_dict[key1]}
@@ -282,7 +291,7 @@ def gtiff2xarray(
 
     with rio.MemoryFile() as memfile:
         memfile.write(r_dict[next(iter(r_dict.keys()))])
-        with memfile.open() as src:
+        with memfile.open(driver=driver) as src:
             r_crs = pyproj.CRS.from_user_input(src.crs)
             if src.nodata is None:
                 try:
@@ -292,12 +301,19 @@ def gtiff2xarray(
             else:
                 nodata = np.dtype(src.dtypes[0]).type(src.nodata)
 
+            ds = xr.open_rasterio(src)
+            valid_dims = list(ds.sizes)
+            if any(d not in valid_dims for d in ds_dims):
+                raise InvalidInputValue("ds_dims", valid_dims)
+
     _geometry = geo2polygon(geometry, geo_crs, r_crs)
+    if not _geometry.is_valid:
+        _geometry = _geometry.buffer(0.0)
 
     def to_dataset(lyr: str) -> xr.DataArray:
         with rio.MemoryFile() as memfile:
             memfile.write(r_dict[lyr])
-            with memfile.open() as src:
+            with memfile.open(driver=driver) as src:
                 geom = [_geometry.intersection(box(*src.bounds))]
                 if geom[0].is_empty:
                     msk, transform, _ = rio_mask.raster_geometry_mask(src, [_geometry], invert=True)
@@ -323,13 +339,12 @@ def gtiff2xarray(
                     coords = {ds_dims[0]: ds.coords[ds_dims[0]], ds_dims[1]: ds.coords[ds_dims[1]]}
                     msk_da = xr.DataArray(msk, coords, dims=ds_dims)
                     ds = ds.where(msk_da, drop=True)
-                    ds.attrs["crs"] = r_crs
+                    ds.attrs["crs"] = r_crs.to_string()
                     ds.name = var_name[lyr]
                     return ds
 
     ds = xr.merge(to_dataset(lyr) for lyr in r_dict.keys())
-    ds.attrs["crs"] = r_crs
-    ds.attrs["transform"] = get_transform(_geometry, ds.sizes[ds_dims[0]], ds.sizes[ds_dims[1]])
+    ds.attrs["crs"] = r_crs.to_string()
 
     if len(ds.variables) - len(ds.dims) == 1:
         ds = ds[list(ds.keys())[0]]
@@ -338,7 +353,7 @@ def gtiff2xarray(
 
 def xarray_geomask(
     ds: Union[xr.Dataset, xr.DataArray],
-    geometry: Union[Polygon, Tuple[float, float, float, float]],
+    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
     geo_crs: str,
     ds_dims: Tuple[str, str] = ("y", "x"),
 ):
@@ -348,7 +363,7 @@ def xarray_geomask(
     ----------
     ds : xarray.Dataset or xarray.DataArray
         The dataset(array) to be masked
-    geometry : Polygon or tuple of length 4
+    geometry : Polygon, MultiPolygon, or tuple of length 4
         The geometry or bounding box to mask the data
     geo_crs : str
         The spatial reference of the input geometry
@@ -361,15 +376,8 @@ def xarray_geomask(
     xarray.Dataset or xarray.DataArray
         The input dataset with a mask applied (np.nan)
     """
-    if not isinstance(ds_dims, tuple) or len(ds_dims) != 2:
-        raise InvalidInputType("ds_dims", "tuple of length 2", '("y", "x")')
-
-    if "crs" not in ds.attrs:
-        raise ValueError("The input dataset is missing the crs attribute.")
-
+    transform, width, height = get_transform(ds, ds_dims)
     _geometry = geo2polygon(geometry, geo_crs, ds.crs)
-    height, width = ds.sizes[ds_dims[0]], ds.sizes[ds_dims[1]]
-    transform = get_transform(_geometry, height, width)
 
     _mask = rio_features.geometry_mask([_geometry], (height, width), transform, invert=True)
 
@@ -384,36 +392,50 @@ def xarray_geomask(
 
 
 def get_transform(
-    geometry: Union[Polygon, Tuple[float, float, float, float]], width: int, height: int
-) -> affine.Affine:
+    ds: Union[xr.Dataset, xr.DataArray],
+    ds_dims: Tuple[str, str] = ("y", "x"),
+) -> Tuple[affine.Affine, int, int]:
     """Get transform of a Polygon or bounding box.
 
     Parameters
     ----------
-    geometry : Polygon or tuple of float
-        The geometry or the bounding box, (west, south, east, north).
-    width: int
-        The width of the target raster in pixels.
-    height: int
-        The height of the target raster in pixels.
+    ds : xarray.Dataset or xarray.DataArray
+        The dataset(array) to be masked
 
     Returns
     -------
-    affin.Affine
-        The affine transform of the geometry.
+    affine.Affine, int, int
+        The affine transform, width, and height
     """
-    if isinstance(geometry, Polygon):
-        west, south, east, north = geometry.bounds
-    elif isinstance(geometry, tuple) and len(geometry) == 4:
-        west, south, east, north = geometry
-    else:
-        raise InvalidInputType("geometry", "Polygon or (west, south, east, north)")
+    if "crs" not in ds.attrs:
+        raise MissingAttribute("crs", ds.attrs.keys())
 
-    return rio_transform.from_bounds(west, south, east, north, width, height)
+    if not isinstance(ds_dims, tuple) or len(ds_dims) != 2:
+        raise InvalidInputType("ds_dims", "tuple of length 2", '("y", "x")')
+
+    ydim, xdim = ds_dims
+    if ydim not in ds.sizes or xdim not in ds.sizes:
+        raise MissingAttribute("input dims", ds.sizes)
+
+    height, width = ds.sizes[ydim], ds.sizes[xdim]
+
+    left, right = ds[xdim].min().item(), ds[xdim].max().item()
+    bottom, top = ds[ydim].min().item(), ds[ydim].max().item()
+
+    x_res = abs(left - right) / (width - 1)
+    y_res = abs(top - bottom) / (height - 1)
+
+    left -= x_res * 0.5
+    right += x_res * 0.5
+    top += y_res * 0.5
+    bottom -= y_res * 0.5
+
+    transform = rio_transform.from_bounds(left, bottom, right, top, width, height)
+    return transform, width, height
 
 
 class MatchCRS:
-    """Match CRS of a input geometry (Polygon, bbox, coord) with the output CRS.
+    """Match CRS of an input geometry (Polygon, bbox, coord) with the output CRS.
 
     Parameters
     ----------
@@ -427,8 +449,9 @@ class MatchCRS:
 
     @staticmethod
     def geometry(geom: Polygon, in_crs: str, out_crs: str) -> Polygon:
-        if not isinstance(geom, Polygon):
-            raise InvalidInputType("geom", "Polygon")
+        """Transform a Polygon."""
+        if not isinstance(geom, (Polygon, MultiPolygon)):
+            raise InvalidInputType("geom", "Polygon or MultiPolygon")
 
         project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
         return ops.transform(project, geom)
@@ -437,6 +460,7 @@ class MatchCRS:
     def bounds(
         geom: Tuple[float, float, float, float], in_crs: str, out_crs: str
     ) -> Tuple[float, float, float, float]:
+        """Transform a bounding box ``(west, south, east, north)``."""
         if not isinstance(geom, tuple) and len(geom) != 4:
             raise InvalidInputType("geom", "tuple of length 4", "(west, south, east, north)")
 
@@ -447,6 +471,7 @@ class MatchCRS:
     def coords(
         geom: Tuple[Tuple[float, ...], Tuple[float, ...]], in_crs: str, out_crs: str
     ) -> Tuple[Any, ...]:
+        """Transform a set of coordinates in form of ((xs), (ys))."""
         if not isinstance(geom, tuple) and len(geom) != 2:
             raise InvalidInputType("geom", "tuple of length 2", "((xs), (ys))")
 
@@ -461,7 +486,9 @@ def check_bbox(bbox: Tuple[float, float, float, float]) -> None:
 
 
 def geo2polygon(
-    geometry: Union[Polygon, Tuple[float, float, float, float]], geo_crs: str, crs: str
+    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
+    geo_crs: str,
+    crs: str,
 ) -> Polygon:
     """Convert a geometry to a Shapely's Polygon and transform to any CRS.
 
@@ -482,7 +509,7 @@ def geo2polygon(
     if isinstance(geometry, tuple):
         return box(*MatchCRS.bounds(geometry, geo_crs, crs))  # type: ignore
 
-    if isinstance(geometry, Polygon):
+    if isinstance(geometry, (Polygon, MultiPolygon)):
         return MatchCRS.geometry(geometry, geo_crs, crs)
 
-    raise InvalidInputType("geometry", "Polygon or tuple of length 4")
+    raise InvalidInputType("geometry", "Polygon, MultiPolygon, or tuple of length 4")
